/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenix6/configs/Configs.hpp"
#include "ctre/phoenix6/controls/ControlRequest.hpp"
#include "ctre/phoenix6/networking/interfaces/Control_Interface.h"
#include <sstream>
#include <units/current.h>
#include <units/dimensionless.h>
#include <units/frequency.h>
#include <units/time.h>


namespace ctre {
namespace phoenix6 {
namespace controls {

/**
 * This feature requires Phoenix Pro.
 *
 * Request a specified motor current (field oriented control).
 *
 * This control request will drive the motor to the requested motor (stator) current value.  This leverages
 * field oriented control (FOC), which means greater peak power than what is documented.  This scales to
 * torque based on Motor's kT constant.
 */
class TorqueCurrentFOC : public ControlRequest
{
    bool ApplyConfigsOnRequest;

    ctre::phoenix::StatusCode SendRequest(const char *network, uint32_t deviceHash, bool cancelOtherRequests, std::shared_ptr<ControlRequest> &req)
    {
        std::stringstream ss;
        auto& ref = requestReference.GetNameValues();
        ss << Output.to<double>(); ref["Output"] = ss.str(); ss.str(std::string());
        ss << MaxAbsDutyCycle.to<double>(); ref["MaxAbsDutyCycle"] = ss.str(); ss.str(std::string());
        ss << Deadband.to<double>(); ref["Deadband"] = ss.str(); ss.str(std::string());
        ss << OverrideCoastDurNeutral; ref["OverrideCoastDurNeutral"] = ss.str(); ss.str(std::string());

        if (req.get() != this)
        {
            auto const reqCast = dynamic_cast<TorqueCurrentFOC *>(req.get());
            if (reqCast != nullptr)
            {
                *reqCast = *this;
            }
            else
            {
                req = std::make_shared<TorqueCurrentFOC>(*this);
            }
        }

        
        std::string strs{ss.str()};
        c_ctre_phoenix6_requestConfigApply(network, deviceHash, ConfigTimeout.to<double>(), strs.c_str(), strs.length(), ApplyConfigsOnRequest);
        ApplyConfigsOnRequest = false;
        return c_ctre_phoenix6_RequestControlTorqueCurrentFOC(network, deviceHash, UpdateFreqHz.to<double>(), cancelOtherRequests, Output.to<double>(), MaxAbsDutyCycle.to<double>(), Deadband.to<double>(), OverrideCoastDurNeutral);
    }

public:
    /**
     * Amount of motor current in Amperes
     */
    units::current::ampere_t Output;
    /**
     * The maximum absolute motor output that can be applied, which effectively
     * limits the velocity. For example, 0.50 means no more than 50% output in
     * either direction.  This is useful for preventing the motor from spinning to
     * its terminal velocity when there is no external torque applied unto the
     * rotor.  Note this is absolute maximum, so the value should be between zero
     * and one.
     */
    units::dimensionless::scalar_t MaxAbsDutyCycle;
    /**
     * Deadband in Amperes.  If torque request is within deadband, the bridge output
     * is neutral. If deadband is set to zero then there is effectively no deadband.
     * Note if deadband is zero, a free spinning motor will spin for quite a while
     * as the firmware attempts to hold the motor's bemf. If user expects motor to
     * cease spinning quickly with a demand of zero, we recommend a deadband of one
     * Ampere. This value will be converted to an integral value of amps.
     */
    units::current::ampere_t Deadband;
    /**
     * Set to true to coast the rotor when output is zero (or within deadband).  Set
     * to false to use the NeutralMode configuration setting (default). This flag
     * exists to provide the fundamental behavior of this control when output is
     * zero, which is to provide 0A (zero torque).
     */
    bool OverrideCoastDurNeutral;

    

    /**
     * \brief The timeout when sending configs associated with this control
     */
    units::time::second_t ConfigTimeout{0.1_s};

    /**
     * \brief The period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     */
    units::frequency::hertz_t UpdateFreqHz{100_Hz}; // Default to 100_Hz

    /**
     *\brief  This feature requires Phoenix Pro.
     *       
     *        Request a specified motor
     *        current (field oriented control).
     *
     *\details  This control request will drive the motor to the requested motor
     *          (stator) current value.  This leverages field oriented control
     *          (FOC), which means greater peak power than what is documented. 
     *          This scales to torque based on Motor's kT constant.
     *
     * \param Output     Amount of motor current in Amperes
     * \param MaxAbsDutyCycle     The maximum absolute motor output that can be
     *                            applied, which effectively limits the
     *                            velocity. For example, 0.50 means no more than
     *                            50% output in either direction.  This is
     *                            useful for preventing the motor from spinning
     *                            to its terminal velocity when there is no
     *                            external torque applied unto the rotor.  Note
     *                            this is absolute maximum, so the value should
     *                            be between zero and one.
     * \param Deadband     Deadband in Amperes.  If torque request is within
     *                     deadband, the bridge output is neutral. If deadband
     *                     is set to zero then there is effectively no deadband.
     *                     Note if deadband is zero, a free spinning motor will
     *                     spin for quite a while as the firmware attempts to
     *                     hold the motor's bemf. If user expects motor to cease
     *                     spinning quickly with a demand of zero, we recommend
     *                     a deadband of one Ampere. This value will be
     *                     converted to an integral value of amps.
     * \param OverrideCoastDurNeutral     Set to true to coast the rotor when
     *                                    output is zero (or within deadband). 
     *                                    Set to false to use the NeutralMode
     *                                    configuration setting (default). This
     *                                    flag exists to provide the fundamental
     *                                    behavior of this control when output
     *                                    is zero, which is to provide 0A (zero
     *                                    torque).
     */
    TorqueCurrentFOC(units::current::ampere_t Output, units::dimensionless::scalar_t MaxAbsDutyCycle, units::current::ampere_t Deadband, bool OverrideCoastDurNeutral) : ControlRequest{"TorqueCurrentFOC"}, ApplyConfigsOnRequest{false}
    {
        this->Output = Output;
        this->MaxAbsDutyCycle = MaxAbsDutyCycle;
        this->Deadband = Deadband;
        this->OverrideCoastDurNeutral = OverrideCoastDurNeutral;
    }

    /**
     *\brief  This feature requires Phoenix Pro.
     *       
     *        Request a specified motor
     *        current (field oriented control).
     *
     *\details  This control request will drive the motor to the requested motor
     *          (stator) current value.  This leverages field oriented control
     *          (FOC), which means greater peak power than what is documented. 
     *          This scales to torque based on Motor's kT constant.
     *
     * \param Output     Amount of motor current in Amperes
     */
    TorqueCurrentFOC(units::current::ampere_t Output) : TorqueCurrentFOC{Output, 1.0, 0.0_A, false}
    {}
    
    /**
     * \brief Modifies this Control Request's Output parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newOutput Parameter to modify
     * \returns Itself
     */
    TorqueCurrentFOC& WithOutput(units::current::ampere_t newOutput)
    {
        Output = newOutput;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's MaxAbsDutyCycle parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newMaxAbsDutyCycle Parameter to modify
     * \returns Itself
     */
    TorqueCurrentFOC& WithMaxAbsDutyCycle(units::dimensionless::scalar_t newMaxAbsDutyCycle)
    {
        MaxAbsDutyCycle = newMaxAbsDutyCycle;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's Deadband parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newDeadband Parameter to modify
     * \returns Itself
     */
    TorqueCurrentFOC& WithDeadband(units::current::ampere_t newDeadband)
    {
        Deadband = newDeadband;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's OverrideCoastDurNeutral parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newOverrideCoastDurNeutral Parameter to modify
     * \returns Itself
     */
    TorqueCurrentFOC& WithOverrideCoastDurNeutral(bool newOverrideCoastDurNeutral)
    {
        OverrideCoastDurNeutral = newOverrideCoastDurNeutral;
        return *this;
    }
    /**
     * \brief Sets the period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     *
     * \param newUpdateFreqHz Parameter to modify
     * \returns Itself
     */
    TorqueCurrentFOC &WithUpdateFreqHz(units::frequency::hertz_t newUpdateFreqHz)
    {
        UpdateFreqHz = newUpdateFreqHz;
        return *this;
    }
    /**
     * Returns a string representation of the object.
     *
     * \returns a string representation of the object.
     */
    std::string ToString() const
    {
        std::stringstream ss;
        ss << "class: TorqueCurrentFOC" << std::endl;
        ss << "Output: " << Output.to<double>() << std::endl;
        ss << "MaxAbsDutyCycle: " << MaxAbsDutyCycle.to<double>() << std::endl;
        ss << "Deadband: " << Deadband.to<double>() << std::endl;
        ss << "OverrideCoastDurNeutral: " << OverrideCoastDurNeutral << std::endl;
        return ss.str();
    }

    /**
     * \brief Forces configs to be applied the next time this is used in a setControl.
     *        This is not necessary in the majority of cases, because Phoenix will make sure configs are
     *        properly set when they are not already set
     */
    void ForceApplyConfigs() { ApplyConfigsOnRequest = true; }
};

}
}
}

